"""Provide support to Lexicon for Subreg.cz DNS changes."""

from __future__ import absolute_import

import logging

import collections

from lexicon.providers.base import Provider as BaseProvider

try:
    import zeep # Optional dependency
except:
    pass

logger = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ['subreg.cz']

def ProviderParser(subparser):
    subparser.add_argument("--auth-username", help="specify username for authentication")
    subparser.add_argument("--auth-password", help="specify password for authentication")

class Provider(BaseProvider):
    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.ssid = None

        client = zeep.Client("https://subreg.cz/wsdl")
        self.api = client.service

    # Authenticate against provider,
    # Make any requests required to get the domain's id for this provider, so it can be used in subsequent calls.
    # Should throw an error if authentication fails for any reason, of if the domain does not exist.
    def authenticate(self):
        """Logs-in the user and checks the domain name"""
        if not self._get_provider_option('auth_username') or not self._get_provider_option('auth_password'):
            raise Exception('No valid authentication data passed, expected: auth-username and auth-password')
        response = self._request_login(self._get_provider_option('auth_username'),
                                       self._get_provider_option('auth_password'))
        if 'ssid' in response:
            self.ssid = response['ssid']
            domains = self.domains_list()
            if any((domain['name'] == self.domain for domain in domains)):
                self.domain_id = self.domain
            else:
                raise Exception("Unknown domain {}".format(self.domain))
        else:
            raise Exception("No SSID provided by server")

    # Create record. If record already exists with the same content, do nothing.
    def create_record(self, type, name, content):
        """Creates a new unique record"""
        found = self.list_records(type, name, content)
        if len(found) > 0:
            return True

        record = self._create_request_record(None, type, name, content,
                                             self._get_lexicon_option('ttl'),
                                             self._get_lexicon_option('priority'))

        self._request_add_dns_record(record)
        return True

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def list_records(self, type=None, name=None, content=None):
        """Lists all records by the type, name and content"""
        return self._list_records(identifier=None, type=type, name=name, content=content)

    # Update a record. Identifier must be specified.
    def update_record(self, identifier, type=None, name=None, content=None):
        """Updates a record. Name changes are allowed, but the record identifier will change"""
        if identifier is not None:
            if name is not None:
                records = self._list_records(identifier=identifier)
                if len(records) == 1 and records[0]['name'] != self._full_name(name):
                    # API does not allow us to update name directly
                    self._update_record_with_name(records[0], type, name, content)
                else:
                    self._update_record(identifier, type, content)
            else:
                self._update_record(identifier, type, content)
        else:
            guessed_record = self._guess_record(type, name)
            self._update_record(guessed_record['id'], type, content)
        return True

    def _update_record(self, identifier, type, content):
        """Updates existing record with no sub-domain name changes"""
        record = self._create_request_record(identifier, type, None, content,
                                             self._get_lexicon_option('ttl'),
                                             self._get_lexicon_option('priority'))

        self._request_modify_dns_record(record)

    def _update_record_with_name(self, old_record, type, new_name, content):
        """Updates existing record and changes it's sub-domain name"""
        new_type = type if type else old_record['type']

        new_ttl = self._get_lexicon_option('ttl')
        if new_ttl is None and 'ttl' in old_record:
            new_ttl = old_record['ttl']

        new_priority = self._get_lexicon_option('priority')
        if new_priority is None and 'priority' in old_record:
            new_priority = old_record['priority']

        new_content = content
        if new_content is None and 'content' in old_record:
            new_content = old_record['content']

        record = self._create_request_record(None,
                                             new_type,
                                             new_name,
                                             new_content,
                                             new_ttl,
                                             new_priority)

        # This will be a different domain name, so no name collision should
        # happen. First create a new entry and when it succeeds, delete the old
        # one.
        self._request_add_dns_record(record)
        self._request_delete_dns_record_by_id(old_record['id'])

    # Delete an existing record.
    # If record does not exist, do nothing.
    # If an identifier is specified, use it, otherwise do a lookup using type, name and content.
    def delete_record(self, identifier=None, type=None, name=None, content=None):
        """Deletes an existing record"""
        to_delete_ids = list()
        if identifier:
            to_delete_ids.append(identifier)
        else:
            for record in self.list_records(type, name, content):
                to_delete_ids.append(record["id"])

        for to_delete_id in to_delete_ids:
            self._request_delete_dns_record_by_id(to_delete_id)
        return True

    # Get list of registered domains
    def domains_list(self):
        response = self._request_domains_list()
        return response['domains'] if 'domains' in response else list()

    def _create_request_record(self, identifier, type, name, content, ttl, priority):
        """Creates record for Subreg API calls"""
        record = collections.OrderedDict()

        # Mandatory content

        # Just for update - not for creation
        if identifier is not None:
            record['id'] = identifier

        record['type'] = type

        # Just for creation - not for update
        if name is not None:
            record['name'] = self._relative_name(name)

        # Optional content
        if content is not None:
            record['content'] = content
        if ttl is not None:
            record['ttl'] = ttl
        if priority is not None:
            record['prio'] = priority
        return record

    def _create_response_record(self, response):
        """Creates record for lexicon API calls"""
        record = dict()
        record['id'] = response['id']
        record['type'] = response['type']
        record['name'] = self._full_name(response['name'])
        if 'content' in response:
            record['content'] = response['content'] or ""
        if 'ttl' in response:
            record['ttl'] = response['ttl']
        if 'prio' in response:
            record['priority'] = response['prio']
        return record

    def _full_name(self, name):
        """Returns full domain name of a sub-domain name"""
        # Handle None and empty strings
        if not name:
            return self.domain
        else:
            return super(Provider, self)._full_name(name)

    def _relative_name(self, name):
        """Returns sub-domain of a domain name"""
        # Handle None and empty strings as None
        if not name:
            return None
        else:
            subdomain = super(Provider, self)._relative_name(name)
            return subdomain if subdomain else None

    # List all records. Return an empty list if no records found
    # identifier, type, name and content are used to filter records.
    def _list_records(self, identifier=None, type=None, name=None, content=None):
        """Lists all records by the specified criteria"""
        response = self._request_get_dns_zone()
        if 'records' in response:
            # Interpret empty string as None because zeep does so too
            content_check = content if content != "" else None
            name_check = self._relative_name(name)

            # Stringize the identifier to prevent any type differences
            identifier_check = str(identifier) if identifier is not None else None

            filtered_records = [record for record in response['records'] if
                        (identifier is None or str(record['id']) == identifier_check) and
                        (type is None or record['type'] == type) and
                        (name is None or record['name'] == name_check) and
                        (content is None or ('content' in record and record['content'] == content_check))]
            records = [self._create_response_record(filtered_record) for filtered_record in filtered_records]
        else:
            records = []
        return records

    def _guess_record(self, type, name=None, content=None):
        """Tries to find existing unique record by type, name and content"""
        records = self._list_records(identifier=None, type=type, name=name, content=content)
        if len(records) == 1:
            return records[0]
        elif len(records) > 1:
            raise Exception('Identifier was not provided and several existing records match the request for {0}/{1}'.format(type,name))
        else:
            raise Exception('Identifier was not provided and no existing records match the request for {0}/{1}'.format(type,name))

    def _request_login(self, login, password):
        """Sends Login request"""
        return self._request("Login",
                             login=login,
                             password=password)

    def _request_domains_list(self):
        """Sends Domains_List request"""
        return self._request("Domains_List")

    def _request_get_dns_zone(self):
        """Sends Get_DNS_Zone request"""
        return self._request("Get_DNS_Zone",
                             domain=self.domain)

    def _request_add_dns_record(self, record):
        """Sends Add_DNS_Record request"""
        return self._request("Add_DNS_Record",
                             domain=self.domain,
                             record=record)

    def _request_modify_dns_record(self, record):
        """Sends Modify_DNS_Record request"""
        return self._request("Modify_DNS_Record",
                             domain=self.domain,
                             record=record)

    def _request_delete_dns_record_by_id(self, identifier):
        """Sends Delete_DNS_Record request"""
        return self._request("Delete_DNS_Record",
                             domain=self.domain,
                             record={ 'id': identifier })

    def _request(self, command, **kwargs):
        """Make request parse response"""
        args = dict(kwargs)
        if self.ssid:
            args['ssid'] = self.ssid
        method = getattr(self.api, command)
        response = method(**args)
        self.response = response
        if response and 'status' in response:
            if response['status'] == 'error':
                raise SubregError(
                    message=response['error']['errormsg'],
                    major=response['error']['errorcode']['major'],
                    minor=response['error']['errorcode']['minor']
                )
            if response['status'] == 'ok':
                return response['data'] if 'data' in response else dict()
            else:
                raise Exception("Invalid status found in SOAP response")
        raise Exception('Invalid response')

class SubregError(Exception):
    def __init__(self, major, minor, message):
        self.major = int(major)
        self.minor = int(minor)
        self.message = message

    def __str__(self):
        return 'Major: {} Minor: {} Message: {}'.format(self.major, self.minor, self.message)



# The Constellix API has some limitations.  We try to paper over them here, but here's what you need to be
# aware of:
#
#  1) SOA records are not first-class record types in the Constellix API, so are not supported.
#  2) We expect all records to use the "Standard" record type, so failover, pools or round robin with
#     failover are not supported.
#  3) Because Constellix represents record sets as a single record with multiple values attached, not as 
#     a set of separate records, create and delete operations end up becoming read/update operations when
#     working with record sets.
#
#     Since these aren't atomic operations, it creates a small window where you could have data loss
#     if multiple processes were trying to work with the same record set.
#
#     This is unlikely to be a problem in most scenarios, but the possilbity is there.  I've reached
#     out to the Constellix folks to see if they have plans to clean up the API to resolve this.

from __future__ import absolute_import

import base64
import contextlib
import hmac
import json
import locale
import logging
import time
import hashlib

import requests
from builtins import bytes

from lexicon.providers.base import Provider as BaseProvider

logger = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ['constellix.com']

def ProviderParser(subparser):
    subparser.add_argument("--auth-username", help="specify the API key username for authentication")
    subparser.add_argument("--auth-token", help="specify secret key for authenticate=")

class Provider(BaseProvider):

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = 'https://api.dns.constellix.com/v1'

    def authenticate(self):
        try:
            payload = self._get('/domains/')
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 404:
                payload = {}
            else:
                raise e
    
        for domain in payload:
            if domain['name'] == self.domain:
                self.domain_id = domain['id']
                self.domain_details = domain
                continue

        if not self.domain_id:
            raise Exception('No domain found')


    # Create record. If record already exists with the same content, do nothing'
    def create_record(self, type, name, content):
        record = {
            'name': self._relative_name(name),
            'ttl': self._get_lexicon_option('ttl'),
            'roundRobin':
                [{'disableFlag': False,
                 'value': content}],
        }
        payload = {}

        try:
            payload = self._post('/domains/{0}/records/{1}/'.format(self.domain_id, type), record)
        except requests.exceptions.HTTPError as e:
            # If there is already a record with that name, we need to do an update.
            if e.response.status_code == 400:
                existing_records = self.list_records(type=type, name=name)
                new_content = [r['content'] for r in existing_records]

                # Only do the update if we are creating a record that doesn't already exist, otherwise
                # Constellix will throw an error.
                if content not in new_content:
                    new_content.append(content)
                    self.update_record(existing_records[0]['id'], type=type, name=name, content=new_content)
            else:
                raise
        logger.debug('create_record: %s', 'name' in payload)
        return True


    # Currently returns the first value for hosts where there may be multiple
    # values.  Need to check to see how this is handled for other providers.
    def list_records(self, type=None, name=None, content=None, identifier=None):
        self._check_type(type)

        # Oddly, Constellix supports API-level filtering for everything except LOC
        # records, so we need to retrieve all records for LOC and filter based on type
        # on our end.
        if not type or type == 'LOC':
            payload = self._get('/domains/{0}/records/'.format(self.domain_id))
        else:
            payload = self._get('/domains/{0}/records/{1}/'.format(self.domain_id, type))

        records = []

        for record in payload:
            for rr in record['roundRobin']:
                processed_record = {
                    'type': record['type'],
                    'name': '{0}.{1}'.format(record['name'], self.domain),
                    'ttl': record['ttl'],
                    'content': rr['value'],
                    'id': record['id']
                    }

                processed_record = self._clean_TXT_record(processed_record)
                records.append(processed_record)

        records = self._filter_records(records, type=type, name=name, content=content, identifier=identifier)

        logger.debug('list_records: %s', records)
        return records

    # Create or update a record.
    def update_record(self, identifier, type=None, name=None, content=None):
        self._check_type(type)

        if content and not isinstance(content, (list)):
            content = [content]

        if identifier and (not type or not name):
            record = self.list_records(identifier=identifier)
            type = record[0]['type']
            name = record[0]['name']
        elif not identifier:
            record = self.list_records(type, name)
            identifier = record[0]['id']

        if not identifier:
            raise Exception("No identifier provided")

        data = {
            'id': identifier,
            'ttl': self._get_lexicon_option('ttl'),
            'name': self._relative_name(name)
        }

        data['roundRobin'] = []

        for c in content:
            data['roundRobin'].append({'disableFlag': False,
                                       'value': c})

        payload = self._put('/domains/{0}/records/{1}/{2}/'.format(self.domain_id, type, identifier), data)

        logger.debug('update_record: %s', True)
        return True

    # Delete an existing record.
    # If record does not exist, do nothing.
    def delete_record(self, identifier=None, type=None, name=None, content=None):
        self._check_type(type)

        records = self.list_records(identifier=identifier, type=type, name=name)

        # If we are filtering delete records by content and we are going to have
        # at least one record left over after deleting, then this becomes an
        # update operation.
        if content:
            current_content = set(r['content'] for r in records)
            if content in current_content and len(current_content) > 1:
                current_content.remove(content)
                self.update_record(records[0]['id'], type=type, name=name, content=list(current_content))
                return True

        delete_record_id = set(record['id'] for record in records)

        # We need a type to do a delete, so pull one from the first record if it's not supplied.
        if not type:
            type = records[0]['type']
        
        for record_id in delete_record_id:
            payload = self._delete('/domains/{0}/records/{1}/{2}/'.format(self.domain_id, type, record_id))

        # is always True at this point, if a non 200 response is returned an error is raised.
        logger.debug('delete_record: %s', True)
        return True

    # Helpers
    def _check_type(self, type=None):
        # Constellix doesn't treat SOA as a separate record type, so we bail on SOA modificiations.
        # It looks like it would be possible to fake SOA CRUD, so an area for possible future
        # improvement

        if type == 'SOA':
            raise Exception('{0} record type is not supported in the Constellix Provider'.format(type))

        return True


    def _filter_records(self, records, type=None, name=None, content=None, identifier=None):
        _records = []
        for record in records:
            if (not identifier or record['id'] == identifier) and \
               (not type or record['type'] == type) and \
               (not name or record['name'] == self._full_name(name)) and \
               (not content or record['content'] == content):
                _records.append(record)
        return _records

    def _request(self, action='GET',  url='/', data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}
        default_headers = {
            'Accept': 'application/json',
            'Content-Type': 'application/json',
            'x-cnsdns-apiKey': self._get_provider_option('auth_username'),
        }
        default_auth = None

        # Date string in epoch format
        request_date = str(int(time.time() * 1000))

        hashed = hmac.new(self._get_provider_option('auth_token').encode('utf-8'), request_date.encode('utf-8'), digestmod=hashlib.sha1)

        default_headers['x-cnsdns-requestDate'] = request_date
        default_headers['x-cnsdns-hmac'] = base64.b64encode(hashed.digest())

        r = requests.request(action, self.api_endpoint + url, params=query_params,
                             data=json.dumps(data),
                             headers=default_headers,
                             auth=default_auth)
        r.raise_for_status()  # if the request fails for any reason, throw an error.

        # PUT and DELETE actions dont return valid json.
        if action == 'DELETE' or action == 'PUT':
            return r.text

        return r.json()
